;;; backup-dir.el --- customization of location of backup files

;; Copyright (C) 1998, 2000, 2002 Free Software Foundation, Inc.

;; Author:  Dave Love <fx@gnu.org>
;; Created: 1998-09-20
;; Keywords: extensions
;; $Revision: 1.1 $

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Commentary:

;; This is a proposed addition for Emacs 20.5 to provide more
;; convenient customization of non-numeric backups, particularly
;; putting them in a single directory fairly uniquely without actually
;; redefining `make-backup-file-name'.  Such functionality is fairly
;; often requested.

;;; Code:

(defcustom make-backup-file-name nil
  "A function to use instead of the default `make-backup-file-name'.
A value of nil indicates the default `make-backup-file-name' behaviour.
This could be buffer-local to do something special for for specific
files.  See also `backup-directory'."
  :group 'backup
  :type '(choice (const :tag "Default" nil)
                 (function :tag "Your function")))

(defcustom backup-directory nil
  "A directory name in which to place all backup files.

If not nil, this should be an existing directory which
`make-backup-file-name' will use, so that backups get made in that
directory.  The file names in this directory will be the full name of
the file backed up with all directory separators changed to `|' so
that backups for different files will not normally clash.  This will
not work correctly if your filesystem truncates the resulting name.

On MS-DOS filesystems without long names, the value of
`backup-directory' is always ignored.

If `backup-directory' is nil, backups for a file are made in that
file's directory as normal."
  :group 'backup
  :type '(choice (const :tag "File's original directory" nil)
                 (directory :tag "Single directory")))

(defcustom backup-directory-alist nil
  "Alist of filename patterns v. backup directory names.
Each element looks like (REGEXP . DIRECTORY)."
  :group 'backup
  :type '(repeat (cons (regexp :tag "Regexp macthing filename")
                       (directory :tag "Directory name"))))

(defcustom backup-replaced-drive-separator "@"
  "String with which the \":\" on pc-based file systems is replaced"
  :type 'string
  :group 'backup)

(defcustom backup-replaced-dir-separator "@"
  "String with which the directory separator is replaced"
  :type 'string
  :group 'backup)

(require 'ginel-string)

(defun make-backup-file-name (file)
  "Create the non-numeric backup file name for FILE.
This is a separate function so you can redefine it for customization.

Built-in customization hooks are provided as follows.

If the variable `make-backup-file-name' is non-nil, it should hold a
function which will be called with FILE as its argument and the
resulting name returned.

Otherwise a match for FILE is sought in `backup-directory-alist' and
used as the value of `backup-directory'.

If the value of the variable `backup-directory' is non-nil, the
name returned is that of a file in the directory named by the
variable.  The non-directory part of the name is derived from
FILE with `directory-sep-char' replaced by
`backup-replaced-dir-separator' so that the names are (normally)
unique. On pc-based file systems the \":\" is replaced with
`backup-replaced-drive-separator'"
  (if make-backup-file-name
      (funcall make-backup-file-name file)
    (let ((backup-directory backup-directory)
          (backup-directory-alist backup-directory-alist))
      (while backup-directory-alist
        (if (string-match (caar backup-directory-alist) file)
            (progn
              (setq backup-directory (cdar backup-directory-alist))
              (setq backup-directory-alist nil)))
        (setq backup-directory-alist (cdr backup-directory-alist)))
      (if (and (eq system-type 'ms-dos)
               (not (msdos-long-file-names)))
          (let ((fn (file-name-nondirectory file)))
            (concat (file-name-directory file)
                    (or (and (string-match "\\`[^.]+\\'" fn)
                             (concat (match-string 0 fn) ".~"))
                        (and (string-match "\\`[^.]+\\.\\(..?\\)?" fn)
                             (concat (match-string 0 fn) "~")))))
        (if backup-directory
            (progn
              (message "%s %s" "dir:" (expand-file-name backup-directory))
              (message "%s %s" "file:" (expand-file-name file))
              (message "%s %s" "backup:" (concat (expand-file-name backup-directory)
                    (string directory-sep-char)
                    (ginel-string-replace 
                     (ginel-string-replace file ":" 
                                           backup-replaced-drive-separator)
                     "/" "@")
                    "~"))
              (concat (expand-file-name backup-directory)
                      (string directory-sep-char)
                      (ginel-string-replace 
                       (ginel-string-replace file ":" 
                                             backup-replaced-drive-separator)
                       "/" "@")
                      "~"))
          (concat file "~"))))))

;;; backup-dir.el ends here
