;; @(#) jpack.el -- Java package statement generator
;; @(#) $Id: jpack.el,v 1.9 1999-02-01 12:42:22+01 ebat311 Exp ebat311 $

;; This file is not part of Emacs

;; Copyright (C) 1998 by David Ponce
;; Author:       David Ponce david.ponce@wanadoo.fr
;; Maintainer:   David Ponce david.ponce@wanadoo.fr
;; Created:      September 28 1998

;; LCD Archive Entry:
;; <el>|David Ponce|david.ponce@wanadoo.fr|
;; <docum>|
;; <date>|$Revision: 1.9 $|~/misc/|

;; COPYRIGHT NOTICE
;;
;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Description:
;;
;;  This package automatically generates a Java package statement. It requires
;;  the JDE package installed and customized to work.
;;  The package name is deducted from the current classpath setting of JDE. When
;;  a directory found in classpath is a root of the current buffer default
;;  directory, the relative path of the default directory from the classpath one
;;  becomes the package name by substituting directory separators by '.'.
;;
;;  For example:
;;      My JDE classpath setting is ("~/java" "/jdk1.1.6/lib/classes.zip")
;;      I edit the file ~/java/FR/test/MyClass.java
;;      The package name generated will be `FR.test'.
;;
;;      My JDE classpath setting is ("~/java" "~/java/test" "/jdk1.1.6/lib/classes.zip")
;;      I edit the file ~/java/test/MyClass.java
;;      "// Default package used." will be generated because the default package can be used.
;;
;;      My JDE classpath setting is ("~/java" "/jdk1.1.6/lib/classes.zip")
;;      I edit the file /usr/java/MyClass.java
;;      No package name will be generated because the directory /usr/java is not
;;      accessible from classpath.

;;; Installation:
;;
;;  Put this file on your Emacs-Lisp load path and add following into your
;;  ~/.emacs startup file
;;
;;      (require 'jpack)

;;; Usage:
;;
;;  M-x `jpack-package-update' to update the Java package statement or insert a new
;;                             at top of buffer.
;;  The function `jpack-generate-package-statement' could be used in template to
;;  automatically generate the package statement.

;;; Customization:
;;
;;  M-x `jpack-customize' to customize all the jpack options.
;;
;;  The following variables could be set:
;;
;;  o `jpack-load-hook'
;;         hook run when package has been loaded.
;;
;;  o `jpack-package-comment'
;;         Java line comment appended to the generated package statement.
;;         Default to " // Generated package name"
;;
;;  o `jpack-default-package-comment'
;;         Java line comment generated when the default package is used.
;;         Default to "// Default package used"
;;
;;  o `jpack-search-classpath-variables'
;;         The list of variables where to search the current classpath list.
;;         Default to '(jde-compile-option-classpath jde-global-classpath)

;;; Support:
;;
;;  Any comments, suggestions, bug reports or upgrade requests are welcome.
;;  Please send them to David Ponce at david.ponce@wanadoo.fr.
;;
;;  This version of jpack was developed with NTEmacs 20.3.1 under MS Windows
;;  NT 4 WKS SP3 and also tested with Emacs 20.3 under Sun Solaris 2.5.
;;  Please, let me know if it works with other OS and versions of Emacs.

;;; Code:

(require 'jde)

(defconst jpack-version "$Revision: 1.9 $"
  "jpack version number.")

(defconst jpack-unknow-package-name
  "*unknow*"
  "The string returned when a package name cant be generated.")

(defconst jpack-package-regexp
  "package .*;.*$"
  "The regexp used to find the Java package statement.")

(defgroup jpack nil
  "jpack package customization"
  :group 'tools
  :prefix "jpack-")

(defcustom jpack-load-hook nil
  "*Hook run when package has been loaded."
  :group 'jpack
  :type 'hook
  )

(defcustom jpack-package-comment " // Generated package name"
  "*Java line comment appended to the generated package statement.
An empty string suppress the generation of this comment."
  :group 'jpack
  :type 'string
  )

(defcustom jpack-default-package-comment "// Default package used"
  "*Java line comment generated when the default package is used.
An empty string suppress the generation of this comment."
  :group 'jpack
  :type 'string
  )

(defcustom jpack-search-classpath-variables
  '(jde-compile-option-classpath jde-global-classpath)
  "*Specify the variables where to search the current classpath list.
The first one which has a non nil value will be used by jpack."
  :group 'jpack
  :type '(repeat variable)
  )

(defun jpack-customize ()
  "Customization of the group jpack."
  (interactive)
  (customize-group "jpack"))

(defun jpack-version-number ()
  "Returns jpack version number."
  (string-match "[0123456789.]+" jpack-version)
  (match-string 0 jpack-version))

(defun jpack-display-version ()
  "Displays jpack version."
  (interactive)
  (message "Using 'jpack' version %s." (jpack-version-number)))

(defun jpack-fullpath (path)
  "Returns the full path of the given path.
~ (HOME) and environment variables references are expanded."
  (expand-file-name (substitute-in-file-name path)))

(defun jpack-get-classpath ()
  "Returns the current classpath list. That is to say the first non-nil value
found in the variables given by `jpack-search-classpath-variables'."
  (let ((search-in jpack-search-classpath-variables)
        (classpath))
    (catch 'jpack-classpath
      (while search-in
        (setq classpath (symbol-value (car search-in)))
        (if classpath
            (throw 'jpack-classpath classpath))
        (setq search-in (cdr search-in))))))

(defun jpack-get-directories-in-classpath ()
  "Returns the list of directories found in classpath."
  (mapcan '(lambda (path)
             (unless (string= path ".") ; "." is ignored in classpath
               (let ((path (jpack-fullpath path)))
                 (when (file-directory-p path)
                   (list (file-name-as-directory path))))))
          (jpack-get-classpath)))

(defun jpack-get-current-directory ()
  "Returns the full path of the current directory."
  (jpack-fullpath default-directory))

(defun jpack-seach-package-directories ()
  "Returns a list of package directory candidates or nil if none found."
  (let ((dir (jpack-get-current-directory))
        (case-fold-search (eq system-type 'windows-nt))) ; case-insensitive for Windows
    (mapcan '(lambda (root)
               (let ((root (regexp-quote root)))
                 (message "Seaching %S in %S..." root dir)
                 (and (string-match root dir)
                      (list (substring dir (match-end 0))))))
            (jpack-get-directories-in-classpath))))

(defun jpack-get-package-directory ()
  "Returns the package directory found or `jpack-unknow-package-name'
if none found."
  (or (jpack-best-package-candidate (jpack-seach-package-directories))
      jpack-unknow-package-name))

(defun jpack-best-package-candidate (candidates)
  "Returns the best package directory candidate from the given list of
package directories. The best is the shortest one that could be found."
  (car (sort candidates '(lambda (dir1 dir2) (string-match (regexp-quote dir1) dir2)))))

(defun jpack-convert-directory-to-package (dir)
  "Converts the given package directory to a valid Java package name
by replacing '/' by '.' and removing extra '/' at end."
  (mapconcat 'identity
             (split-string dir (char-to-string directory-sep-char))
             "."))

(defun jpack-generate-package-statement ()
  "Generates a Java package statement. If the package name cant be generated
this function returns an empty string."
  (let ((package-name (jpack-get-package-directory)))
    (cond ((string= package-name jpack-unknow-package-name)
           (message "The current directory is not accessible from classpath.")
           "")
          ((string= package-name "")
           (message "Default package used.")
           jpack-default-package-comment)
          (t (message "package %s;%s"
                      (jpack-convert-directory-to-package package-name)
                      jpack-package-comment)))))

(defun jpack-package-update ()
  "Updates or replaces the Java package statement in the current buffer.
If the package statement does not exist a new one is inserted at the
top of the buffer. If the package name cant be generated nothing is done.
The major mode of current buffer must be 'jde-mode'."
  (interactive)
  (if (eq major-mode 'jde-mode)
      (let ((package (jpack-generate-package-statement)))
        (unless (string= package "")
          (goto-char (point-min))
          (if (re-search-forward jpack-package-regexp nil t)
              (replace-match package)
            (progn
              (insert package)
              (newline)))))
    (message "Invalid major mode found. Must be 'jde-mode'.")))

(provide 'jpack)
(run-hooks 'jpack-load-hook)

;;; Change History:

;;
;; $Log: jpack.el,v $
;; Revision 1.9  1999-02-01 12:42:22+01  ebat311
;; Can now customize classpath source and generated comments.
;; Follows a remark from Phillip Lord <P.Lord@mdx.ac.uk>.
;;
;; Revision 1.8  1998/10/21 16:04:48  ebat311
;; Because file names are case-insensitive for Windows `jpack-seach-package-directories' now sets `case-fold-search' to t if `system-type' is windows-nt.
;;
;; Revision 1.7  1998/10/20 10:04:39  ebat311
;; Have run `untabify' on the whole source (follows a remark from ricky@siemensdc.com).
;;
;; Revision 1.6  1998/10/15 09:28:14  ebat311
;; Function `jpack-convert-directory-to-package' now uses
;; `directory-sep-char' instead of hard coded "/".
;;
;; Revision 1.5  1998/10/05 21:15:44  ebat311
;; Some coding optimized and simplified.
;;
;; Revision 1.5  1998/09/29 22:35:43  ebat311
;; Outdated (renamed to 'C:\RCS\C\users\dpe\emacs\jpack.el,v').
;;
;; Revision 1.4  1998/09/29 09:05:43  ebat311
;; New version-number management functions.
;;
;; Revision 1.3  1998/09/28 22:50:20  ebat311
;; Copyright notice updated.
;;
;; Revision 1.2  1998/09/28 22:07:02  ebat311
;; Fixed an error in function `jpack-version-number' which did not display
;; the version number.
;;
;; Revision 1.1  1998/09/28 21:59:29  ebat311
;; Initial revision
;;
;;

;;; jpack.el ends here.
